@@ -18,6 +18,7 @@
 WARNING: This module does not mlock() secrets; your private keys may end up on
 disk in swap! Use with caution!
 """
+
 import hmac
 import struct
 import ctypes
@@ -152,17 +153,20 @@ _libsecp256k1.secp256k1_ecdsa_sign.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
 _libsecp256k1.secp256k1_ecdsa_signature_serialize_der.restype = ctypes.c_int
 _libsecp256k1.secp256k1_ecdsa_signature_serialize_der.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_size_t), ctypes.c_char_p]
 
-_libsecp256k1.secp256k1_ecdsa_sign_recoverable.restype = ctypes.c_int
-_libsecp256k1.secp256k1_ecdsa_sign_recoverable.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p, ctypes.c_void_p]
+_libsecp256k1_has_pubkey_recovery = False
+if getattr(_libsecp256k1, 'secp256k1_ecdsa_sign_recoverable', None):
+    _libsecp256k1_has_pubkey_recovery = True
+    _libsecp256k1.secp256k1_ecdsa_sign_recoverable.restype = ctypes.c_int
+    _libsecp256k1.secp256k1_ecdsa_sign_recoverable.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p, ctypes.c_void_p]
 
-_libsecp256k1.secp256k1_ecdsa_recoverable_signature_serialize_compact.restype = ctypes.c_int
-_libsecp256k1.secp256k1_ecdsa_recoverable_signature_serialize_compact.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_int), ctypes.c_char_p]
+    _libsecp256k1.secp256k1_ecdsa_recoverable_signature_serialize_compact.restype = ctypes.c_int
+    _libsecp256k1.secp256k1_ecdsa_recoverable_signature_serialize_compact.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_int), ctypes.c_char_p]
 
-_libsecp256k1.secp256k1_ecdsa_recover.restype = ctypes.c_int
-_libsecp256k1.secp256k1_ecdsa_recover.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
+    _libsecp256k1.secp256k1_ecdsa_recover.restype = ctypes.c_int
+    _libsecp256k1.secp256k1_ecdsa_recover.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
 
-_libsecp256k1.secp256k1_ecdsa_recoverable_signature_parse_compact.restype = ctypes.c_int
-_libsecp256k1.secp256k1_ecdsa_recoverable_signature_parse_compact.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int]
+    _libsecp256k1.secp256k1_ecdsa_recoverable_signature_parse_compact.restype = ctypes.c_int
+    _libsecp256k1.secp256k1_ecdsa_recoverable_signature_parse_compact.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int]
 
 _libsecp256k1.secp256k1_ec_pubkey_serialize.restype = ctypes.c_int
 _libsecp256k1.secp256k1_ec_pubkey_serialize.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_size_t), ctypes.c_char_p, ctypes.c_uint]
@@ -202,28 +206,26 @@ _libsecp256k1_seed = urandom(32)
 assert(_libsecp256k1.secp256k1_context_randomize(_libsecp256k1_context_sign, _libsecp256k1_seed) == 1)
 
 
-# We do not want to make CKey a subclass of bytes, as we do with CPubKey,
-# to prevent accidental leak of key data when it is included in debug logs, etc.
-# to access key data, you need to explicitly say key.secret
-class CKey(object):
+class CKeyMixin():
     """An encapsulated private key
 
     Attributes:
 
     pub           - The corresponding CPubKey for this private key
-    secret        - Raw secret bytes of private key
+    secret_bytes  - Secret data, 32 bytes (needed because subclasses may have trailing data)
 
     is_compressed - True if compressed
 
     """
-    def __init__(self, secret, compressed=True):
-        assert len(secret) == 32
-        self.secret = secret
 
+    def __init__(self, s, compressed=True):
         raw_pubkey = ctypes.create_string_buffer(64)
 
+        # no need for explicit secp256k1_ec_seckey_verify()
+        # because secp256k1_ec_pubkey_create() will do
+        # the same checks and ensure that secret data is valid
         result = _libsecp256k1.secp256k1_ec_pubkey_create(
-            _libsecp256k1_context_sign, raw_pubkey, secret)
+            _libsecp256k1_context_sign, raw_pubkey, self.secret_bytes)
 
         if result != 1:
             raise ValueError('Invalid private key data')
@@ -234,9 +236,9 @@ class CKey(object):
     def is_compressed(self):
         return self.pub.is_compressed
 
-    def set_compressed(self, compressed):
-        raw_pubkey = self.pub._to_raw()
-        self.pub = CPubKey._from_raw(raw_pubkey, compressed=compressed)
+    @property
+    def secret_bytes(self):
+        return self[:32]
 
     def sign(self, hash):
         if not isinstance(hash, bytes):
@@ -246,10 +248,10 @@ class CKey(object):
 
         raw_sig = ctypes.create_string_buffer(64)
         result = _libsecp256k1.secp256k1_ecdsa_sign(
-            _libsecp256k1_context_sign, raw_sig, hash, self, None, None)
+            _libsecp256k1_context_sign, raw_sig, hash, self.secret_bytes, None, None)
         assert 1 == result
         sig_size0 = ctypes.c_size_t()
-        sig_size0.value = 75
+        sig_size0.value = SIGNATURE_SIZE
         mb_sig = ctypes.create_string_buffer(sig_size0.value)
         result = _libsecp256k1.secp256k1_ecdsa_signature_serialize_der(
             _libsecp256k1_context_sign, mb_sig, ctypes.byref(sig_size0), raw_sig)
@@ -264,10 +266,14 @@ class CKey(object):
         if len(hash) != 32:
             raise ValueError('Hash must be exactly 32 bytes long')
 
+        if not _libsecp256k1_has_pubkey_recovery:
+            raise RuntimeError('libsecp256k1 compiled without pubkey recovery functions. '
+                               'sign_compact is not functional.')
+
         recoverable_sig = ctypes.create_string_buffer(COMPACT_SIGNATURE_SIZE)
 
         result = _libsecp256k1.secp256k1_ecdsa_sign_recoverable(
-            _libsecp256k1_context_sign, recoverable_sig, hash, self, None, None)
+            _libsecp256k1_context_sign, recoverable_sig, hash, self.secret_bytes, None, None)
 
         assert 1 == result
 
@@ -288,6 +294,16 @@ class CKey(object):
         return self.pub.verify_nonstrict(hash, sig)
 
 
+class CKey(bytes, CKeyMixin):
+    "Standalone privkey class"
+
+    @classmethod
+    def from_secret_bytes(cls, secret, compressed=True):
+        assert compressed
+        assert len(secret) == 32
+        return cls(secret)
+
+
 class CPubKey(bytes):
     """An encapsulated public key
 
@@ -318,6 +334,8 @@ class CPubKey(bytes):
 
     @classmethod
     def _from_raw(cls, raw_pubkey, compressed=True):
+        if len(raw_pubkey) != 64:
+            raise ValueError('raw pubkey must be 64 bytes')
         pub_size0 = ctypes.c_size_t()
         pub_size0.value = PUBLIC_KEY_SIZE
         pub = ctypes.create_string_buffer(pub_size0.value)
@@ -342,6 +360,10 @@ class CPubKey(bytes):
         if len(sig) != COMPACT_SIGNATURE_SIZE:
             raise ValueError("Signature should be %d characters, not [%d]" % (COMPACT_SIGNATURE_SIZE, len(sig)))
 
+        if not _libsecp256k1_has_pubkey_recovery:
+            raise RuntimeError('libsecp256k1 compiled without pubkey recovery functions. '
+                               'recover_compact is not functional.')
+
         recid = (sig[0] - 27) & 3
         compressed = ((sig[0] - 27) & 4) != 0
 
@@ -426,7 +448,9 @@ class CPubKey(bytes):
         # extra encode/decode is wasteful, but the result is that verification
         # is still roughly 4 times faster than with openssl's ECDSA_verify()
         norm_sig = ctypes.c_void_p(0)
-        _ssl.d2i_ECDSA_SIG(ctypes.byref(norm_sig), ctypes.byref(ctypes.c_char_p(sig)), len(sig))
+        result = _ssl.d2i_ECDSA_SIG(ctypes.byref(norm_sig), ctypes.byref(ctypes.c_char_p(sig)), len(sig))
+        if not result:
+            return False
 
         derlen = _ssl.i2d_ECDSA_SIG(norm_sig, 0)
         if derlen == 0:
@@ -434,46 +458,71 @@ class CPubKey(bytes):
             return False
 
         norm_der = ctypes.create_string_buffer(derlen)
-        _ssl.i2d_ECDSA_SIG(norm_sig, ctypes.byref(ctypes.pointer(norm_der)))
+        result = _ssl.i2d_ECDSA_SIG(norm_sig, ctypes.byref(ctypes.pointer(norm_der)))
+
         _ssl.ECDSA_SIG_free(norm_sig)
 
+        if not result:
+            return False
+
         return self.verify(hash, norm_der)
 
 
-class CExtKeyMixin():
+class CExtKeyBase():
 
-    def _parse_buf(self, buf):
-        if len(buf) != 74:
+    def _check_length(self):
+        if len(self) != 74:
             raise ValueError('Invalid length for extended key')
 
-        self.depth = buf[0]
-        self.parent_fp = buf[1:5]
-        self.child_number = struct.unpack(">L", buf[5:9])[0]
-        self.chaincode = buf[9:41]
+    @property
+    def depth(self):
+        return self[0]
 
-        return buf[41:74]
+    @property
+    def parent_fp(self):
+        return self[1:5]
+
+    @property
+    def child_number(self):
+        return struct.unpack(">L", self.child_number_bytes)[0]
+
+    @property
+    def child_number_bytes(self):
+        return self[5:9]
+
+    @property
+    def chaincode(self):
+        return self[9:41]
+
+    @property
+    def key_bytes(self):
+        return self[41:74]
 
 
-class CExtKey(bytes, CExtKeyMixin):
+class CExtKeyMixin(CExtKeyBase):
     """An encapsulated extended private key
 
+    Attributes:
+
     priv          - The corresponding CKey for extended privkey
     pub           - shortcut property for priv.pub
     """
 
-    def __new__(cls, buf):
-        self = super(CExtKey, cls).__new__(cls, buf)
+    def __init__(self, _s):
 
-        # Note that we ignore first byte - for xpubkey,
+        self._check_length()
+
+        # NOTE: we ignore first byte - for xpubkey,
         # this is pubkey prefix byte.
         # For xprivkey, this byte is supposed to be zero,
         # but Bitcoin Core ignores that and do not check.
         # We also do not check, to be compatible.
-        secret_bytes = self._parse_buf(buf)[1:]
+        raw_priv = self.key_bytes[1:]
 
-        self.priv = CKey(secret_bytes)
-
-        return self
+        # NOTE: cannot make self.priv a @property method
+        # because we need to check if the privkey is valid
+        # CKey() will do this for us.
+        self.priv = self._key_class.from_secret_bytes(raw_priv)
 
     @property
     def pub(self):
@@ -489,7 +538,7 @@ class CExtKey(bytes, CExtKeyMixin):
         parent_fp = child_number_packed = b'\x00\x00\x00\x00'
         privkey = hmac_hash[:32]
         chaincode = hmac_hash[32:]
-        return cls(bytes([depth]) + parent_fp + child_number_packed + chaincode + bytes([0]) + privkey)
+        return cls.from_bytes(bytes([depth]) + parent_fp + child_number_packed + chaincode + bytes([0]) + privkey)
 
     def derive(self, child_number):
         if self.depth >= 255:
@@ -507,12 +556,12 @@ class CExtKey(bytes, CExtKeyMixin):
                                   hashlib.sha512).digest()
         else:
             bip32_hash = hmac.new(self.chaincode,
-                                  bytes([0]) + self.priv.secret + child_number_packed,
+                                  bytes([0]) + self.priv.secret_bytes + child_number_packed,
                                   hashlib.sha512).digest()
 
         chaincode = bip32_hash[32:]
 
-        child_privkey = ctypes.create_string_buffer(self.priv.secret, size=32)
+        child_privkey = ctypes.create_string_buffer(self.priv.secret_bytes, size=32)
 
         result = _libsecp256k1.secp256k1_ec_privkey_tweak_add(
             _libsecp256k1_context_sign, child_privkey, bip32_hash)
@@ -522,29 +571,33 @@ class CExtKey(bytes, CExtKeyMixin):
 
         parent_fp = self.pub.key_id[:4]
         cls = self.__class__
-        return cls(bytes([depth]) + parent_fp + child_number_packed + chaincode + bytes([0]) + child_privkey)
+        return cls.from_bytes(bytes([depth]) + parent_fp + child_number_packed + chaincode + bytes([0]) + child_privkey)
 
     def neuter(self):
-        child_number_packed = struct.pack(">L", self.child_number)
-        return CExtPubKey(bytes([self.depth]) + self.parent_fp + child_number_packed
-                          + self.chaincode + self.pub)
+        return self._xpub_class.from_bytes(
+            bytes([self.depth]) + self.parent_fp + self.child_number_bytes + self.chaincode + self.pub)
 
 
-class CExtPubKey(bytes, CExtKeyMixin):
+class CExtPubKeyMixin(CExtKeyBase):
     """An encapsulated extended public key
 
+    Attributes:
+
     pub           - The corresponding CPubKey for extended pubkey
 
     """
 
-    def __new__(cls, buf):
-        self = super(CExtPubKey, cls).__new__(cls, buf)
-        pubkey_bytes = self._parse_buf(buf)
+    def __init__(self, _s):
 
-        self.pub = CPubKey(pubkey_bytes)
+        self._check_length()
+
+        self.pub = CPubKey(self.key_bytes)
         if not self.pub.is_fullyvalid:
             raise ValueError('pubkey part of xpubkey is not valid')
-        return self
+
+    @classmethod
+    def from_bytes(cls, data):
+        return cls(data)
 
     def derive(self, child_number):
         if (child_number >> 31) != 0:
@@ -584,7 +637,7 @@ class CExtPubKey(bytes, CExtKeyMixin):
 
         parent_fp = self.pub.key_id[:4]
         cls = self.__class__
-        return cls(bytes([depth]) + parent_fp + child_number_packed + chaincode + child_pubkey)
+        return cls.from_bytes(bytes([depth]) + parent_fp + child_number_packed + chaincode + child_pubkey)
 
     def __str__(self):
         return repr(self)
@@ -592,8 +645,27 @@ class CExtPubKey(bytes, CExtKeyMixin):
     def __repr__(self):
         return '%s(%s)' % (self.__class__.__name__, super(CExtPubKey, self).__repr__())
 
+
+class CExtPubKey(bytes, CExtPubKeyMixin):
+    "Standalone extended pubkey class"
+
+
+class CExtKey(bytes, CExtKeyMixin):
+    "Standalone extended key class"
+    _key_class = CKey
+    _xpub_class = CExtPubKey
+
+    @classmethod
+    def from_bytes(cls, data):
+        return cls(data)
+
+
 __all__ = (
     'CKey',
     'CPubKey',
-    'CExtPubKey'
+    'CExtKey',
+    'CExtPubKey',
+    'CKeyMixin',
+    'CExtKeyMixin',
+    'CExtPubKeyMixin'
 )
